from dataclasses import dataclass
from typing import Any, Optional

import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder

from cesnet_datazoo.constants import SERVICEMAP_CATEGORY_COLUMN, SERVICEMAP_PROVIDER_COLUMN


@dataclass()
class ClassInfo:
    target_names: list[str]
    num_classes: int
    known_apps: list[str]
    unknown_apps: list[str]
    unknown_class_label: int
    group_matrix: np.ndarray
    has_provider: dict[str, bool]
    provider_mapping: dict[str, str]
    provider_members: dict[str, list[str]]
    categories_mapping: dict[str, Optional[str]]

def create_class_info(servicemap: Any, encoder: LabelEncoder, known_apps_database_enum: dict[int, str], unknown_apps_database_enum: dict[int, str]) -> ClassInfo:
    known_apps = sorted(known_apps_database_enum.values())
    unknown_apps = sorted(unknown_apps_database_enum.values())
    target_names_arr = encoder.classes_
    assert known_apps == list(target_names_arr[:-1])
    group_matrix = np.array([[a == b or
                (a in servicemap.index and b in servicemap.index and
                not pd.isnull(servicemap.loc[a, SERVICEMAP_PROVIDER_COLUMN]) and not pd.isnull(servicemap.loc[b, SERVICEMAP_PROVIDER_COLUMN]) and
                servicemap.loc[a, SERVICEMAP_PROVIDER_COLUMN] == servicemap.loc[b, SERVICEMAP_PROVIDER_COLUMN])
                for a in target_names_arr] for b in target_names_arr])
    has_provider = {app: app in servicemap.index and not pd.isnull(servicemap.loc[app, SERVICEMAP_PROVIDER_COLUMN]) for app in target_names_arr}
    provider_mapping = {app: servicemap.loc[app, SERVICEMAP_PROVIDER_COLUMN] if has_provider[app] else app for app in target_names_arr}
    providers = sorted({provider_mapping[app] for app in target_names_arr if has_provider[app]})
    provider_members = {p: [app for app in target_names_arr if provider_mapping[app] == p] for p in providers}
    categories_mapping = {app: servicemap.loc[app, SERVICEMAP_CATEGORY_COLUMN] if app in servicemap.index else None for app in target_names_arr}
    return ClassInfo(
            target_names=list(target_names_arr),
            num_classes=len(known_apps),
            known_apps=known_apps,
            unknown_apps=unknown_apps,
            unknown_class_label=len(known_apps),
            group_matrix=group_matrix,
            has_provider=has_provider,
            provider_mapping=provider_mapping,
            provider_members=provider_members,
            categories_mapping=categories_mapping,
    )



import pickle
from typing import Any

import yaml


def pickle_load(path: str) -> Any:
    with open(path, "rb") as f:
        loaded_object = pickle.load(f)
    return loaded_object

def pickle_dump(object_to_save: Any, path: str) -> None:
    with open(path, "wb") as f:
        pickle.dump(object_to_save, f, protocol=pickle.HIGHEST_PROTOCOL)

def yaml_load(path: str) -> Any:
    with open(path, "r") as f:
        loaded_object = yaml.load(f, Loader=yaml.SafeLoader)
    return loaded_object

def yaml_dump(object_to_save: Any, path: str) -> None:
    with open(path, "w") as f:
        yaml.dump(object_to_save, f)


import os

import requests
from tqdm import tqdm


def simple_download(url: str, file_path: str):
    r = requests.get(url)
    with open(file_path, "wb") as file:
        file.write(r.content)

def resumable_download(url: str, file_path: str, chunk_size: int = 1024**2, silent: bool = False):
    r1 = requests.get(url, stream=True)
    redirected_url = r1.url
    content_size = int(r1.headers["Content-Length"])

    if os.path.exists(file_path):
        temp_size = os.path.getsize(file_path)
    else:
        temp_size = 0
    if temp_size == content_size:
        return

    headers = {"Range": f"bytes={temp_size}-"}
    r2 = requests.get(redirected_url, stream=True, headers=headers)
    if not silent:
        print(f"File size: {content_size / (1024**3):0.2f}GB")
        print(f"Remaining: {(content_size - temp_size) / (1024**3):0.2f}GB")

    progress_bar = tqdm(total=content_size - temp_size, unit="B", unit_scale=True, unit_divisor=1024, disable=silent)
    with open(file_path, "ab") as file:
        for data in r2.iter_content(chunk_size=chunk_size):
            file.write(data)
            progress_bar.update(len(data))
    progress_bar.close()



from __future__ import annotations

import dataclasses
import hashlib
import json
import os
import warnings
from dataclasses import InitVar, field
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Literal, Optional

import yaml
from pydantic import model_validator
from pydantic.dataclasses import dataclass
from sklearn.preprocessing import MinMaxScaler, RobustScaler, StandardScaler

from cesnet_datazoo.constants import (PHIST_BIN_COUNT, PPI_MAX_LEN, SELECTED_TCP_FLAGS,
                                      TCP_PPI_CHANNELS, UDP_PPI_CHANNELS)

if TYPE_CHECKING:
    from cesnet_datazoo.datasets.cesnet_dataset import CesnetDataset

Scaler = RobustScaler | StandardScaler | MinMaxScaler | None

class ScalerEnum(Enum):
    """Available scalers for flow statistics, packet sizes, and inter-packet times."""
    STANDARD = "standard"
    """Standardize features by removing the mean and scaling to unit variance - [`sklearn.preprocessing.StandardScaler`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html)."""
    ROBUST = "robust"
    """Robust scaling with the median and the interquartile range - [`sklearn.preprocessing.RobustScaler`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.RobustScaler.html)."""
    MINMAX = "minmax"
    """Scaling to a (0, 1) range - [`sklearn.preprocessing.MinMaxScaler`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html)."""
    NO_SCALER = "no-scaler"
    """No scaling."""
    def __str__(self): return self.value

class Protocol(Enum):
    TLS = "TLS"
    QUIC = "QUIC"
    def __str__(self): return self.value

class ValidationApproach(Enum):
    """The validation approach defines which samples should be used for creating a validation set."""
    SPLIT_FROM_TRAIN = "split-from-train"
    """Split train data into train and validation.
    Scikit-learn [`train_test_split`](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html)
    is used to create a random stratified validation set. The fraction of validation samples is defined in `train_val_split_fraction`."""
    VALIDATION_DATES = "validation-dates"
    """Use separate validation dates to create a validation set. Validation dates need to be specified in `val_dates`, and the name of the validation period in `val_period_name`."""
    NO_VALIDATION = "no-validation"
    """Do not use validation. The validation dataloader and dataframe will not be available."""
    def __str__(self): return self.value

class AppSelection(Enum):
    """
    Applications can be divided into *known* and *unknown* classes. To use a dataset in the standard closed-world setting, use `ALL_KNOWN ` to select all the applications as *known*.
    Use `TOPX_KNOWN` or `EXPLICIT_UNKNOWN` for the open-world setting and evaluation of out-of-distribution or open-set recognition methods.
    The `LONGTERM_FIXED` is for long-term measurements when it is desired to use the same applications for multiple subsequent train and test periods.
    """
    ALL_KNOWN = "all-known"
    """Use all applications as *known*."""
    TOPX_KNOWN = "topx-known"
    """Use the first X (`apps_selection_topx`) most frequent (with the most samples) applications as *known*, and the rest as *unknown*.
    Applications with the same provider are never separated, i.e., all applications of a given provider are either *known* or *unknown*."""
    EXPLICIT_UNKNOWN = "explicit-unknown"
    """Use the provided list of applications (`apps_selection_explicit_unknown`) as *unknown*, and the rest as *known*."""
    LONGTERM_FIXED = "longterm-fixed"
    """Use fixed application selection. Provide a tuple of `(known_apps_database_enum, unknown_apps_database_enum)` in `apps_selection_fixed_longterm`."""
    def __str__(self): return self.value

class MinTrainSamplesCheck(Enum):
    """
    Depending on the selected train dates, there might be applications with *not enough* samples for training (what is *not enough* will depend on the selected classification model).
    The threshold for the minimum number of samples can be set with `min_train_samples_per_app`, and its default value is 100.
    With the `DISABLE_APPS ` approach, these applications will be disabled and not used for training or testing.
    With the `WARN_AND_EXIT ` approach, the script will print a warning and exit if applications with not enough samples are encountered.
    To disable this check, set `min_train_samples_per_app` to 0.
    """
    WARN_AND_EXIT = "warn-and-exit"
    """Warn and exit if there are not enough training samples for some applications. It is up to the user to manually add these applications to `disabled_apps`."""
    DISABLE_APPS = "disable-apps"
    """Disable applications with not enough training samples."""
    def __str__(self): return self.value

class DataLoaderOrder(Enum):
    """
    Validation and test sets are always loaded in sequential order — sequential meaning in the order of dates and time.
    However, for the train set, it is sometimes required to iterate it in random order (for example, for training a neural network).
    Thus, use `RANDOM ` if your classification model requires it; `SEQUENTIAL` otherwise.
    This setting affects only [train_dataloader][datasets.cesnet_dataset.CesnetDataset.get_train_dataloader]. Dataframe [get_train_df][datasets.cesnet_dataset.CesnetDataset.get_train_df] is always created in sequential order.
    """
    RANDOM = "random"
    """Iterate train data in random order."""
    SEQUENTIAL = "sequential"
    """Iterate train data in sequential (datetime) order."""
    def __str__(self): return self.value

@dataclass(frozen=True)
class TrainDataParams():
    database_filename: str
    train_period_name: str
    train_tables_paths: list[str]
    apps_selection: AppSelection
    apps_selection_topx: int
    apps_selection_explicit_unknown: list[str]
    apps_selection_fixed_longterm: Optional[tuple[dict[int, str], dict[int, str]]]
    disabled_apps: list[str]
    min_train_samples_check: MinTrainSamplesCheck
    min_train_samples_per_app: int

@dataclass(frozen=True)
class TestDataParams():
    database_filename: str
    test_period_name: str
    test_tables_paths: list[str]
    known_apps_database_enum: dict[int, str]
    unknown_apps_database_enum: dict[int, str]

class C:
    arbitrary_types_allowed = True

@dataclass(config=C)
class DatasetConfig():
    """
    The main class for the configuration of:

    - Train, validation, test sets (dates, sizes, validation approach).
    - Application selection — either the standard closed-world setting (only *known* classes) or the open-world setting (*known* and *unknown* classes).
    - Feature scaling.
    - Dataloader options like batch sizes, order of loading, or number of workers.

    When initializing this class, pass a [`CesnetDataset`][datasets.cesnet_dataset.CesnetDataset] instance to be configured and the desired configuration. Available options are [here][config.DatasetConfig--configuration-options].

    Attributes:
        dataset: The dataset instance to be configured
        data_root: Taken from the dataset instance
        database_filename: Taken from the dataset instance
        database_path: Taken from the dataset instance
        servicemap_path: Taken from the dataset instance
        flowstats_features: Taken from `dataset.metadata.flowstats_features`

    # Configuration options

    Attributes:
        train_period_name: Name of the train period. See [instructions][config.DatasetConfig--how-to-configure-train-validation-and-test-sets].
        train_dates: Dates used for creating a train set.
        train_dates_weigths: To use a non-uniform distribution of samples across train dates.
        val_approach: How a validation set should be created. Either split train data into train and validation, have a separate validation period, or no validation at all. `Default: SPLIT_FROM_TRAIN`
        train_val_split_fraction: The fraction of validation samples when splitting from the train set. `Default: 0.2`
        val_period_name: Name of the validation period. See [instructions][config.DatasetConfig--how-to-configure-train-validation-and-test-sets].
        val_dates: Dates used for creating a validation set.
        no_test_set: Disable the test set. `Default: False`
        test_period_name: Name of the test period. See [instructions][config.DatasetConfig--how-to-configure-train-validation-and-test-sets].
        test_dates: Dates used for creating a test set.

        apps_selection: How to select application classes. `Default: ALL_KNOWN`
        apps_selection_topx: Take top X as known.
        apps_selection_explicit_unknown: Provide a list of unknown applications.
        apps_selection_fixed_longterm: Provide enums of known and unknown applications. This is suitable for long-term measurements.
        disabled_apps: List of applications to be disabled and not used at all.
        min_train_samples_check: How to handle applications with *not enough* training samples. `Default: DISABLE_APPS`
        min_train_samples_per_app: Defines the threshold for *not enough*. `Default: 100`

        random_state: Fix all random processes performed during dataset initialization. `Default: 420`
        fold_id: To perform N-fold cross-validation, set this to `1..N`. Each fold will use the same configuration but a different random seed. `Default: 0`
        train_workers: Number of workers for loading train data. `0` means that the data will be loaded in the main process. `Default: 4`
        test_workers: Number of workers for loading test data. `0` means that the data will be loaded in the main process. `Default: 1`
        val_workers: Number of workers for loading validation data. `0` means that the data will be loaded in the main process. `Default: 1`
        batch_size: Number of samples per batch. `Default: 192`
        test_batch_size: Number of samples per batch for loading validation and test data. `Default: 2048`
        preload_test: Whether to dump the test set with `numpy.savez_compressed` and preload it in future runs. Useful when running a lot of experiments with the same dataset configuration. `Default: False`
        train_size: Size of the train set. See [instructions][config.DatasetConfig--how-to-configure-train-validation-and-test-sets]. `Default: all`
        val_known_size: Size of the validation set. See [instructions][config.DatasetConfig--how-to-configure-train-validation-and-test-sets]. `Default: all`
        test_known_size: Size of the test set. See [instructions][config.DatasetConfig--how-to-configure-train-validation-and-test-sets]. `Default: all`
        val_unknown_size: Size of the unknown classes validation set. Use for evaluation in the open-world setting. `Default: 0`
        test_unknown_size: Size of the unknown classes test set. Use for evaluation in the open-world setting. `Default: 0`
        train_dataloader_order: Whether to load train data in sequential or random order. `Default: RANDOM`
        train_dataloader_seed: Seed for loading train data in random order. `Default: None`

        return_ips: Use for IP-based classification. Dataloaders will return data in this tuple format `((SRC_IP, DST_IP, SRC_PORT, DST_PORT), LABELS)`. Dataframes are not available when this option is used. `Default: False`
        return_torch: Use for returning `torch.Tensor` from dataloaders. Dataframes are not available when this option is used. `Default: False`
        use_packet_histograms: Whether to use packet histogram features, if available in the dataset. `Default: True`
        normalize_packet_histograms: Whether to normalize packet histograms. If true, bins contain fractions instead of absolute numbers. `Default: True`
        use_tcp_features: Whether to use TCP features, if available in the dataset. `Default: True`
        use_push_flags: Whether to use push flags in packet sequences, if available in the dataset. `Default: False`
        zero_ppi_start: Zeroing out the first N packets of each packet sequence. `Default: 0`
        fit_scalers_samples: Fraction of train samples used for fitting feature scalers, if float. The absolute number of samples otherwise. `Default: 0.25`
        flowstats_scaler: Which scaler to use for flow statistics. Options are [`ROBUST`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.RobustScaler.html) | [`STANDARD`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html) | [`MINMAX`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html) | `NO_SCALER`. `Default: ROBUST`
        flowstats_clip: Quantile clip before the scaling of flow statistics. Should limit the influence of outliers. Set to `1` to disable. `Default: 0.99`
        psizes_scaler: Which scaler to use for packet sizes. Options are [`ROBUST`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.RobustScaler.html) | [`STANDARD`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html) | [`MINMAX`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html) | `NO_SCALER`. `Default: STANDARD`
        psizes_max: Max clip packet sizes before scaling. `Default: 1460`
        ipt_scaler: Which scaler to use for inter-packet times. Options are [`ROBUST`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.RobustScaler.html) | [`STANDARD`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html) | [`MINMAX`](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html) | `NO_SCALER`. `Default: STANDARD`
        ipt_min: Min clip inter-packet times before scaling. `Default: 0`
        ipt_max: Max clip inter-packet times before scaling. `Default: 15000`

    # How to configure train, validation, and test sets
    There are three options for how to define train/validation/test dates.

    1. Choose a predefined time period (`train_period_name`, `val_period_name`, or `test_period_name`) available in `dataset.time_periods` and leave the list of dates (`train_dates`, `val_dates`, or `test_dates`) empty.
    2. Provide a list of dates and a name for the time period. The dates are checked against `dataset.available_dates`.
    3. Do not specify anything and use the dataset's defaults `dataset.default_train_period_name` and `dataset.default_test_period_name`.

    There are two options for configuring sizes of train/validation/test sets.

    1. Select an appropriate dataset size (default is `S`) when creating the [`CesnetDataset`][datasets.cesnet_dataset.CesnetDataset] instance and leave `train_size`, `val_known_size`, and `test_known_size` with their default `all` value.
    This will create train/validation/test sets with all samples available in the selected dataset size (of course, depending on the selected dates and validation approach).
    2. Provide exact sizes in `train_size`, `val_known_size`, and `test_known_size`. This will create train/validation/test sets of the given sizes by doing a random subset.
    This is especially useful when using the `ORIG` dataset size and want to control the size of experiments.

    !!! tip Validation set
        The default approach for creating a validation set is to randomly split the train data into train and validation. The second approach is to define separate validation dates. See [ValidationApproach][config.ValidationApproach].

    """
    dataset: InitVar[CesnetDataset]
    data_root: str = field(init=False)
    database_filename: str =  field(init=False)
    database_path: str =  field(init=False)
    servicemap_path: str = field(init=False)
    flowstats_features: list[str] = field(init=False)

    train_period_name: str = ""
    train_dates: list[str] = field(default_factory=list)
    train_dates_weigths: Optional[list[int]] = None
    val_approach: ValidationApproach = ValidationApproach.SPLIT_FROM_TRAIN
    train_val_split_fraction: float = 0.2
    val_period_name: str = ""
    val_dates: list[str] = field(default_factory=list)
    no_test_set: bool = False
    test_period_name: str = ""
    test_dates: list[str] = field(default_factory=list)

    apps_selection: AppSelection = AppSelection.ALL_KNOWN
    apps_selection_topx: int = 0
    apps_selection_explicit_unknown: list[str] = field(default_factory=list)
    apps_selection_fixed_longterm: Optional[tuple[dict[int, str], dict[int, str]]] = None
    disabled_apps: list[str] = field(default_factory=list)
    min_train_samples_check: MinTrainSamplesCheck = MinTrainSamplesCheck.DISABLE_APPS
    min_train_samples_per_app: int = 100

    random_state: int = 420
    fold_id: int = 0
    train_workers: int = 4
    test_workers: int = 1
    val_workers: int = 1
    batch_size: int = 192
    test_batch_size: int = 2048
    preload_test: bool = False
    train_size: int | Literal["all"] = "all"
    val_known_size: int | Literal["all"] = "all"
    test_known_size: int | Literal["all"] = "all"
    val_unknown_size: int | Literal["all"] = 0
    test_unknown_size: int | Literal["all"] = 0
    train_dataloader_order: DataLoaderOrder = DataLoaderOrder.RANDOM
    train_dataloader_seed: Optional[int] = None

    return_ips: bool = False
    return_torch: bool = False
    use_packet_histograms: bool = True
    normalize_packet_histograms: bool = True
    use_tcp_features: bool = True
    use_push_flags: bool = False
    zero_ppi_start: int = 0
    fit_scalers_samples: int | float = 0.25
    flowstats_scaler: ScalerEnum = ScalerEnum.ROBUST
    flowstats_clip: float = 0.99
    psizes_scaler: ScalerEnum = ScalerEnum.STANDARD
    psizes_max: int = 1460
    ipt_scaler: ScalerEnum = ScalerEnum.STANDARD
    ipt_min: int = 0
    ipt_max: int = 15000

    def __post_init__(self, dataset: CesnetDataset):
        """
        Ensures valid configuration. Catches all incompatible options and raise exceptions as soon as possible.
        """
        self.data_root = dataset.data_root
        self.servicemap_path = dataset.servicemap_path
        self.database_filename = dataset.database_filename
        self.database_path = dataset.database_path
        self.flowstats_features = dataset.metadata.flowstats_features

        # Configure train dates
        if len(self.train_dates) > 0 and self.train_period_name == "":
            raise ValueError("train_period_name has to be specified when train_dates are set")
        if len(self.train_dates) == 0 and self.train_period_name != "":
            if self.train_period_name not in dataset.time_periods:
                raise ValueError(f"Unknown train_period_name {self.train_period_name}. Use time period available in dataset.time_periods")
            self.train_dates = dataset.time_periods[self.train_period_name]
        if len(self.train_dates) == 0 and self.test_period_name == "":
            self.train_period_name = dataset.default_train_period_name
            self.train_dates = dataset.time_periods[dataset.default_train_period_name]
        # Configure test dates
        if self.no_test_set:
            if (len(self.test_dates) > 0 or self.test_period_name != ""):
                raise ValueError("test_dates and test_period_name cannot be specified when no_test_set is true")
        else:
            if len(self.test_dates) > 0 and self.test_period_name == "":
                raise ValueError("test_period_name has to be specified when test_dates are set")
            if len(self.test_dates) == 0 and self.test_period_name != "":
                if self.test_period_name not in dataset.time_periods:
                    raise ValueError(f"Unknown test_period_name {self.test_period_name}. Use time period available in dataset.time_periods")
                self.test_dates = dataset.time_periods[self.test_period_name]
            if len(self.test_dates) == 0 and self.test_period_name == "":
                self.test_period_name = dataset.default_test_period_name
                self.test_dates = dataset.time_periods[dataset.default_test_period_name]
        # Configure val dates
        if (self.val_approach == ValidationApproach.NO_VALIDATION or self.val_approach == ValidationApproach.SPLIT_FROM_TRAIN) and (len(self.val_dates) > 0 or self.val_period_name != ""):
            raise ValueError("val_dates and val_period_name cannot be specified when val_approach is no-validation or split-from-train")
        if self.val_approach == ValidationApproach.VALIDATION_DATES:
            if len(self.val_dates) > 0 and self.val_period_name == "":
                raise ValueError("val_period_name has to be specified when val_dates are set")
            if len(self.val_dates) == 0 and self.val_period_name != "":
                if self.val_period_name not in dataset.time_periods:
                    raise ValueError(f"Unknown val_period_name {self.val_period_name}. Use time period available in dataset.time_periods")
                self.val_dates = dataset.time_periods[self.val_period_name]
            if len(self.val_dates) == 0 and self.val_period_name == "":
                raise ValueError("val_period_name and val_dates (or val_period_name from dataset.time_periods) have to be specified when val_approach is validation-dates")
        # Check if train, val, and test dates are available in the dataset
        if dataset.available_dates:
            unknown_train_dates = [t for t in self.train_dates if t not in dataset.available_dates]
            unknown_val_dates = [t for t in self.val_dates if t not in dataset.available_dates]
            unknown_test_dates = [t for t in self.test_dates if t not in dataset.available_dates]
            if len(unknown_train_dates) > 0:
                raise ValueError(f"Unknown train dates {unknown_train_dates}. Use dates available in dataset.available_dates (collection period {dataset.metadata.collection_period})" \
                                + (f". These dates are missing from the dataset collection period {dataset.metadata.missing_dates_in_collection_period}" if dataset.metadata.missing_dates_in_collection_period else ""))
            if len(unknown_val_dates) > 0:
                raise ValueError(f"Unknown validation dates {unknown_val_dates}. Use dates available in dataset.available_dates (collection period {dataset.metadata.collection_period})" \
                                + (f". These dates are missing from the dataset collection period {dataset.metadata.missing_dates_in_collection_period}" if dataset.metadata.missing_dates_in_collection_period else ""))
            if len(unknown_test_dates) > 0:
                raise ValueError(f"Unknown test dates {unknown_test_dates}. Use dates available in dataset.available_dates (collection period {dataset.metadata.collection_period})" \
                                + (f". These dates are missing from the dataset collection period {dataset.metadata.missing_dates_in_collection_period}" if dataset.metadata.missing_dates_in_collection_period else ""))
        # Check time order of train, val, and test periods
        train_dates = [datetime.strptime(date_str, "%Y%m%d").date() for date_str in self.train_dates]
        test_dates = [datetime.strptime(date_str, "%Y%m%d").date() for date_str in self.test_dates]
        if not self.no_test_set and min(test_dates) <= max(train_dates):
            warnings.warn(f"Some test dates ({min(test_dates).strftime('%Y%m%d')}) are before or equal to the last train date ({max(train_dates).strftime('%Y%m%d')}). This might lead to improper evaluation and should be avoided.")
        if self.val_approach == ValidationApproach.VALIDATION_DATES:
            val_dates = [datetime.strptime(date_str, "%Y%m%d").date() for date_str in self.val_dates]
            if min(val_dates) <= max(train_dates):
                warnings.warn(f"Some validation dates ({min(val_dates).strftime('%Y%m%d')}) are before or equal to the last train date ({max(train_dates).strftime('%Y%m%d')}). This might lead to improper evaluation and should be avoided.")
            if not self.no_test_set and min(test_dates) <= max(val_dates):
                warnings.warn(f"Some test dates ({min(test_dates).strftime('%Y%m%d')}) are before or equal to the last validation date ({max(val_dates).strftime('%Y%m%d')}). This might lead to improper evaluation and should be avoided.")
        # Configure features
        if dataset.metadata.protocol == Protocol.TLS and self.use_tcp_features:
            self.flowstats_features = self.flowstats_features + SELECTED_TCP_FLAGS
            if self.use_push_flags and "PUSH_FLAG" not in dataset.metadata.features_in_packet_sequences:
                raise ValueError("This TLS dataset does not support use_push_flags")
        if self.use_packet_histograms:
            if len(dataset.metadata.packet_histogram_features) > 0:
                self.flowstats_features = self.flowstats_features + dataset.metadata.packet_histogram_features
            else:
                self.use_packet_histograms = False
        if dataset.metadata.protocol == Protocol.QUIC:
            self.use_tcp_features = False
            if self.use_push_flags:
                raise ValueError("QUIC datasets do not support use_push_flags")
        # When train_dates_weigths are used, train_size and val_known_size have to be specified
        if self.train_dates_weigths is not None:
            if len(self.train_dates_weigths) != len(self.train_dates):
                raise ValueError("train_dates_weigths has to have the same length as train_dates")
            if self.train_size == "all":
                raise ValueError("train_size cannot be 'all' when train_dates_weigths are speficied")
            if self.val_approach == ValidationApproach.SPLIT_FROM_TRAIN and self.val_known_size == "all":
                raise ValueError("val_known_size cannot be 'all' when train_dates_weigths are speficied and validation_approach is split-from-train")
        # App selection
        if self.apps_selection == AppSelection.ALL_KNOWN:
            self.val_unknown_size = 0
            self.test_unknown_size = 0
            if self.apps_selection_topx != 0 or len(self.apps_selection_explicit_unknown) > 0 or self.apps_selection_fixed_longterm is not None:
                raise ValueError("apps_selection_topx, apps_selection_explicit_unknown, and apps_selection_fixed_longterm cannot be specified when apps_selection is all-known")
        if self.apps_selection == AppSelection.TOPX_KNOWN and self.apps_selection_topx == 0:
            raise ValueError("apps_selection_topx has to be greater than 0 when apps_selection is top-x-known")
        if self.apps_selection == AppSelection.EXPLICIT_UNKNOWN and len(self.apps_selection_explicit_unknown) == 0:
            raise ValueError("apps_selection_explicit_unknown has to be specified when apps_selection is explicit-unknown")
        if self.apps_selection == AppSelection.LONGTERM_FIXED:
            if self.apps_selection_fixed_longterm is None:
                raise ValueError("apps_selection_fixed_longterm, a tuple of (known_apps_database_enum, unknown_apps_database_enum), has to be specified when apps_selection is longterm-fixed")
            if len(self.disabled_apps) > 0:
                raise ValueError("disabled_apps cannot be specified when apps_selection is longterm-fixed")
            if self.min_train_samples_per_app != 0:
                raise ValueError("min_train_samples_per_app has to be 0 when apps_selection is longterm-fixed")
        if sum((self.apps_selection_topx != 0, len(self.apps_selection_explicit_unknown) > 0, self.apps_selection_fixed_longterm is not None)) > 1:
            raise ValueError("apps_selection_topx, apps_selection_explicit_unknown, and apps_selection_fixed_longterm should not be specified at the same time")
        # More asserts
        if self.zero_ppi_start > PPI_MAX_LEN:
            raise ValueError(f"zero_ppi_start has to be <= {PPI_MAX_LEN}")
        if isinstance(self.fit_scalers_samples, float) and (self.fit_scalers_samples <= 0 or self.fit_scalers_samples > 1):
            raise ValueError("fit_scalers_samples has to be either float between 0 and 1 (giving the fraction of training samples used for fitting scalers) or an integer")

    def get_flowstats_features_len(self) -> int:
        """Gets the number of flow statistics features."""
        n = 0
        for f in self.flowstats_features:
            if f.startswith("PHIST_"):
                n += PHIST_BIN_COUNT
            else:
                n += 1
        return n

    def get_flowstats_feature_names_expanded(self) -> list[str]:
        """Gets names of flow statistics features. Packet histograms are expanded into individual features."""
        name_mapping = {
            "PHIST_SRC_SIZES": [f"PSIZE_BIN{i}" for i in range(1, PHIST_BIN_COUNT + 1)],
            "PHIST_DST_SIZES": [f"PSIZE_BIN{i}_REV" for i in range(1, PHIST_BIN_COUNT + 1)],
            "PHIST_SRC_IPT": [f"IPT_BIN{i}" for i in range(1, PHIST_BIN_COUNT + 1)],
            "PHIST_DST_IPT": [f"IPT_BIN{i}_REV" for i in range(1, PHIST_BIN_COUNT + 1)],
            "FLOW_ENDREASON_IDLE": "FEND_IDLE",
            "FLOW_ENDREASON_ACTIVE": "FEND_ACTIVE",
            "FLOW_ENDREASON_END": "FEND_END",
            "FLOW_ENDREASON_OTHER": "FEND_OTHER",
        }
        feature_names = []
        for f in self.flowstats_features:
            if f not in name_mapping:
                f = f if not f.startswith("FLAG") else "F" + f.lstrip("FLAG")
                f = f.upper()
                feature_names.append(f)
            elif isinstance(name_mapping[f], list):
                feature_names.extend(name_mapping[f])
            else:
                feature_names.append(name_mapping[f])
        assert len(feature_names) == self.get_flowstats_features_len()
        return feature_names

    def get_ppi_feature_names(self) -> list[str]:
        """Gets names of flattened per-packet information (PPI) features."""
        ppi_feature_names = [f"IPT_{i}" for i in range(1, PPI_MAX_LEN + 1)] + \
                               [f"DIR_{i}" for i in range(1, PPI_MAX_LEN + 1)] + \
                               [f"SIZE_{i}" for i in range(1, PPI_MAX_LEN + 1)]
        if self.use_push_flags:
            ppi_feature_names += [f"PUSH_{i}" for i in range(1, PPI_MAX_LEN + 1)]
        return ppi_feature_names

    def get_ppi_channels(self) -> int:
        """Gets the number of per-packet information (PPI) channels."""
        if self.use_push_flags:
            return TCP_PPI_CHANNELS
        else:
            return UDP_PPI_CHANNELS

    def get_feature_names(self, flatten_ppi: bool = False) -> list[str]:
        """
        Gets feature names.

        Parameters:
            flatten_ppi: Whether to flatten PPI into individual feature names or keep one `PPI` column.
        """
        feature_names = self.get_ppi_feature_names() if flatten_ppi else ["PPI"]
        feature_names += self.get_flowstats_feature_names_expanded()
        return feature_names

    def _get_train_tables_paths(self) -> list[str]:
        return list(map(lambda t: f"/flows/D{t}", self.train_dates))

    def _get_val_tables_paths(self) -> list[str]:
        if self.val_approach == ValidationApproach.SPLIT_FROM_TRAIN:
            return list(map(lambda t: f"/flows/D{t}", self.train_dates))
        return list(map(lambda t: f"/flows/D{t}", self.val_dates))

    def _get_test_tables_paths(self) -> list[str]:
        return list(map(lambda t: f"/flows/D{t}", self.test_dates))

    def _get_train_data_hash(self) -> str:
        train_data_params = self._get_train_data_params()
        params_hash = hashlib.sha256(json.dumps(dataclasses.asdict(train_data_params), sort_keys=True, default=str).encode()).hexdigest()
        params_hash = params_hash[:10]
        return params_hash

    def _get_train_data_path(self) -> str:
        params_hash = self._get_train_data_hash()
        return os.path.join(self.data_root, "train-data", f"{params_hash}_{self.random_state}", f"fold_{self.fold_id}")

    def _get_train_data_params(self) -> TrainDataParams:
        return TrainDataParams(
            database_filename=self.database_filename,
            train_period_name=self.train_period_name,
            train_tables_paths=self._get_train_tables_paths(),
            apps_selection=self.apps_selection,
            apps_selection_topx=self.apps_selection_topx,
            apps_selection_explicit_unknown=self.apps_selection_explicit_unknown,
            apps_selection_fixed_longterm=self.apps_selection_fixed_longterm,
            disabled_apps=self.disabled_apps,
            min_train_samples_per_app=self.min_train_samples_per_app,
            min_train_samples_check=self.min_train_samples_check,)

    def _get_val_data_params_and_path(self, known_apps_database_enum: dict[int, str], unknown_apps_database_enum: dict[int, str]) -> tuple[TestDataParams, str]:
        assert self.val_approach == ValidationApproach.VALIDATION_DATES
        val_data_params = TestDataParams(
            database_filename=self.database_filename,
            test_period_name=self.val_period_name,
            test_tables_paths=self._get_val_tables_paths(),
            known_apps_database_enum=known_apps_database_enum,
            unknown_apps_database_enum=unknown_apps_database_enum,)
        params_hash = hashlib.sha256(json.dumps(dataclasses.asdict(val_data_params), sort_keys=True).encode()).hexdigest()
        params_hash = params_hash[:10]
        val_data_path = os.path.join(self.data_root, "val-data", f"{params_hash}_{self.random_state}")
        return val_data_params, val_data_path

    def _get_test_data_params_and_path(self, known_apps_database_enum: dict[int, str], unknown_apps_database_enum: dict[int, str]) -> tuple[TestDataParams, str]:
        test_data_params = TestDataParams(
            database_filename=self.database_filename,
            test_period_name=self.test_period_name,
            test_tables_paths=self._get_test_tables_paths(),
            known_apps_database_enum=known_apps_database_enum,
            unknown_apps_database_enum=unknown_apps_database_enum,)
        params_hash = hashlib.sha256(json.dumps(dataclasses.asdict(test_data_params), sort_keys=True).encode()).hexdigest()
        params_hash = params_hash[:10]
        test_data_path = os.path.join(self.data_root, "test-data", f"{params_hash}_{self.random_state}")
        return test_data_params, test_data_path

    @model_validator(mode="before")
    @classmethod
    def check_deprecated_args(cls, values):
        kwargs = values.kwargs
        if "train_period" in kwargs:
            warnings.warn("train_period is deprecated. Use train_period_name instead.")
            kwargs["train_period_name"] = kwargs["train_period"]
        if "val_period" in kwargs:
            warnings.warn("val_period is deprecated. Use val_period_name instead.")
            kwargs["val_period_name"] = kwargs["val_period"]
        if "test_period" in kwargs:
            warnings.warn("test_period is deprecated. Use test_period_name instead.")
            kwargs["test_period_name"] = kwargs["test_period"]
        return values

    def __str__(self):
        _process_tag = yaml.emitter.Emitter.process_tag
        _ignore_aliases = yaml.Dumper.ignore_aliases
        yaml.emitter.Emitter.process_tag = lambda self, *args, **kw: None
        yaml.Dumper.ignore_aliases = lambda self, *args, **kw: True
        s = yaml.dump(dataclasses.asdict(self), sort_keys=False)
        yaml.emitter.Emitter.process_tag = _process_tag
        yaml.Dumper.ignore_aliases = _ignore_aliases
        return s




import numpy as np
import pandas as pd
from torch.utils.data import DataLoader
from tqdm import tqdm

from cesnet_datazoo.constants import APP_COLUMN


def load_from_dataloader(dataloader: DataLoader, silent: bool = False) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    data_ppi = []
    data_flowstats = []
    labels = []
    if not silent:
        print("Loading data from dataloader")
    for batch_ppi, batch_flowstats, batch_labels in tqdm(dataloader, total=len(dataloader), disable=silent):
        data_ppi.append(batch_ppi)
        data_flowstats.append(batch_flowstats)
        labels.append(batch_labels)
    data_ppi = np.concatenate(data_ppi)
    data_flowstats = np.concatenate(data_flowstats)
    labels = np.concatenate(labels)
    return data_ppi, data_flowstats, labels

def create_df_from_dataloader(dataloader: DataLoader, feature_names: list[str], flatten_ppi: bool = False, silent: bool = False) -> pd.DataFrame:
    data_ppi, data_flowstats, labels = load_from_dataloader(dataloader, silent=silent)
    if flatten_ppi:
        data_ppi = data_ppi.reshape(data_ppi.shape[0], -1)
        data = np.column_stack((data_ppi, data_flowstats))
        df = pd.DataFrame(data=data, columns=feature_names)
    else:
        ppi_column, *feature_names = feature_names
        df = pd.DataFrame(data=data_flowstats, columns=feature_names)
        df.insert(0, column=ppi_column, value=list(data_ppi))
    df[APP_COLUMN] = labels
    return df



DATASET_SIZES = {
    "XS": 10_000_000,
    "S": 25_000_000,
    "M": 50_000_000,
    "L": 100_000_000,
}

# Per-packet information (PPI) constants
IPT_POS = 0
DIR_POS = 1
SIZE_POS = 2
PUSH_FLAGS_POS = 3
PPI_MAX_LEN = 30
TCP_PPI_CHANNELS = 4
UDP_PPI_CHANNELS = 3

# Features
FLOWSTATS_TO_SCALE =  ["BYTES", "BYTES_REV", "PACKETS", "PACKETS_REV", "PPI_LEN", "PPI_ROUNDTRIPS", "PPI_DURATION", "DURATION"]
FLOWSTATS_NO_CLIP = ["PPI_LEN", "PPI_ROUNDTRIPS", "PPI_DURATION", "DURATION"]
SELECTED_TCP_FLAGS = ["FLAG_CWR", "FLAG_CWR_REV", "FLAG_ECE", "FLAG_ECE_REV", "FLAG_PSH_REV", "FLAG_RST", "FLAG_RST_REV", "FLAG_FIN", "FLAG_FIN_REV"]
FLOWEND_REASON_FEATURES = ["FLOW_ENDREASON_IDLE", "FLOW_ENDREASON_ACTIVE", "FLOW_ENDREASON_END", "FLOW_ENDREASON_OTHER"]
IP_FEATURES = ["SRC_IP", "DST_IP", "SRC_PORT", "DST_PORT"]
PHISTS_FEATURES = ["PHIST_SRC_SIZES", "PHIST_DST_SIZES", "PHIST_SRC_IPT", "PHIST_DST_IPT"]
PHIST_BIN_COUNT = 8

# Column names
APP_COLUMN = "APP"
CATEGORY_COLUMN = "CATEGORY"
PPI_COLUMN = "PPI"
TLS_SNI_COLUMN = "TLS_SNI"
QUIC_SNI_COLUMN = "QUIC_SNI"

# Servicemap constants
SERVICEMAP_PROVIDER_COLUMN = "Service Provider"
SERVICEMAP_CATEGORY_COLUMN = "Service Category"
SERVICEMAP_FILE = "servicemap.csv"

# Class labels
UNKNOWN_STR_LABEL = "_unknown"
DEFAULT_BACKGROUND_CLASS = "default-background"
GOOGLE_BACKGROUND_CLASS = "google-background"

# Indices
INDICES_TABLE_POS = 0
INDICES_INDEX_POS = 1
INDICES_LABEL_POS = 2



import atexit
import logging
import os
import time
import warnings
from typing import Any, Optional

import numpy as np
import pandas as pd
import tables as tb
import torch
from numpy.lib.recfunctions import drop_fields, structured_to_unstructured
from sklearn.preprocessing import MinMaxScaler, RobustScaler, StandardScaler, LabelEncoder
from torch.utils.data import Dataset
from typing_extensions import assert_never

from cesnet_datazoo.config import (AppSelection, DatasetConfig, MinTrainSamplesCheck, ScalerEnum,
                                   TestDataParams, TrainDataParams, Scaler)
from cesnet_datazoo.constants import (APP_COLUMN, CATEGORY_COLUMN, DIR_POS, FLOWSTATS_NO_CLIP,
                                      FLOWSTATS_TO_SCALE, IPT_POS, INDICES_INDEX_POS,
                                      INDICES_LABEL_POS, INDICES_TABLE_POS, IP_FEATURES,
                                      PHIST_BIN_COUNT, PHISTS_FEATURES, PPI_COLUMN, SIZE_POS,
                                      UNKNOWN_STR_LABEL)
from cesnet_datazoo.pytables_data.apps_split import (is_background_app,
                                                     split_apps_topx_with_provider_groups)
from cesnet_datazoo.utils.fileutils import pickle_dump, pickle_load
from cesnet_datazoo.utils.random import RandomizedSection, get_fresh_random_generator

log = logging.getLogger(__name__)


class PyTablesDataset(Dataset):
    def __init__(self, database_path: str, tables_paths: list[str], indices: Optional[np.ndarray], flowstats_features: list[str], disabled_apps: Optional[list[str]] = None, preload: bool = False, preload_blob: Optional[str] = None, return_ips: bool = False, return_all_fields: bool = False):
        self.database_path = database_path
        self.tables_paths = tables_paths
        self.tables = {}
        self.flowstats_features = flowstats_features
        self.preload = preload
        self.preload_blob = preload_blob
        self.return_all_fields = return_all_fields
        self.return_ips = return_ips
        if indices is None:
            self.set_all_indices(disabled_apps=disabled_apps)
        else:
            self.indices = indices

    def __getitem__(self, batch_idx):
        # log.debug(f"worker {self.worker_id}: __getitem__")
        if self.preload:
            batch_data = self.data[batch_idx]
        else:
            batch_data = load_data_from_pytables(tables=self.tables, indices=self.indices[batch_idx], data_dtype=self.data_dtype)
        if self.return_all_fields:
            return (batch_data, batch_idx)
        return_data = (batch_data[PPI_COLUMN].astype("float32"), batch_data[self.flowstats_features], list(map(self.app_enum, batch_data[APP_COLUMN])))
        if self.return_ips:
            return_data = (batch_data[IP_FEATURES],) + return_data
        return return_data

    def __len__(self):
        return len(self.indices)

    def pytables_worker_init(self, worker_id=0):
        self.worker_id = worker_id
        log.debug(f"Initializing dataloader worker id {self.worker_id}")
        self.database, self.tables = load_database(database_path=self.database_path, tables_paths=self.tables_paths)
        atexit.register(self.cleanup)
        self.app_enum = self.tables[0].get_enum(APP_COLUMN)
        self.cat_enum = self.tables[0].get_enum(CATEGORY_COLUMN)
        self.data_dtype = self.tables[0].dtype
        if self.preload:
            data = None
            if self.preload_blob and os.path.isfile(self.preload_blob):
                try:
                    data = np.load(self.preload_blob)["data"]
                    log.info(f"Found dataset blob to preload: {self.preload_blob}")
                except:
                    pass # ignore if the file is corrupted (or being written at the moment)
            if data is None:
                data = load_data_from_pytables(tables=self.tables, indices=self.indices, data_dtype=self.data_dtype)
            self.data = data
            if self.preload_blob and not os.path.isfile(self.preload_blob):
                np.savez_compressed(self.preload_blob, data=self.data)
        log.debug(f"Finish initialization worker id {self.worker_id}")

    def get_app_enum(self) -> tb.Enum:
        if self.app_enum:
            return self.app_enum
        database, tables = load_database(database_path=self.database_path, tables_paths=self.tables_paths)
        app_enum = tables[0].get_enum(APP_COLUMN)
        cat_enum = tables[0].get_enum(CATEGORY_COLUMN)
        self.app_enum, self.cat_enum = app_enum, cat_enum
        database.close()
        return app_enum

    def get_cat_enum(self) -> tb.Enum:
        if self.cat_enum:
            return self.cat_enum
        database, tables = load_database(database_path=self.database_path, tables_paths=self.tables_paths)
        app_enum = tables[0].get_enum(APP_COLUMN)
        cat_enum = tables[0].get_enum(CATEGORY_COLUMN)
        self.app_enum, self.cat_enum = app_enum, cat_enum
        database.close()
        return cat_enum

    def set_all_indices(self, disabled_apps: Optional[list[str]] = None):
        """
        This should be called from the main process, before dataloader workers split the work.
        Does no filter apps with not enough samples.
        """
        database, tables = load_database(database_path=self.database_path, tables_paths=self.tables_paths)
        app_enum = tables[0].get_enum(APP_COLUMN)
        disabled_apps_ids = list(map(lambda x: app_enum[x], disabled_apps)) if disabled_apps is not None else []
        base_labels = {}
        base_indices = {}
        for i in range(len(tables)):
            base_labels[i] = tables[i].read(field=APP_COLUMN)
            base_indices[i] = np.nonzero(np.isin(base_labels[i], disabled_apps_ids, invert=True))[0]
        indices = np.column_stack((
            np.concatenate([[table_id] * len(base_indices[table_id]) for table_id in tables]),
            np.concatenate(list(base_indices.values())),
            np.concatenate(list(base_labels.values()))
        )).astype(np.int32)
        self.indices = indices
        database.close()

    def cleanup(self):
        self.database.close()

def worker_init_fn(worker_id):
    worker_info = torch.utils.data.get_worker_info() # type: ignore
    dataset = worker_info.dataset
    dataset.pytables_worker_init(worker_id)

def pytables_ip_collate_fn(batch):
    ips, _, _, labels = batch
    return ips, labels

def pytables_collate_fn(batch: tuple, 
                        flowstats_scaler: Scaler, flowstats_quantiles: pd.Series,
                        psizes_scaler: Scaler, psizes_max: int,
                        ipt_scaler: Scaler, ipt_min: int, ipt_max: int,
                        use_push_flags: bool, use_packet_histograms: bool, normalize_packet_histograms: bool, zero_ppi_start: int,
                        encoder: LabelEncoder, known_apps: list[str], return_torch: bool = False):
    x_ppi, x_flowstats, labels = batch
    x_ppi = x_ppi.transpose(0, 2, 1)
    orig_shape = x_ppi.shape
    ppi_channels = x_ppi.shape[-1]
    x_ppi = x_ppi.reshape(-1, ppi_channels)
    x_ppi[:, IPT_POS] = x_ppi[:, IPT_POS].clip(max=ipt_max, min=ipt_min)
    x_ppi[:, SIZE_POS] = x_ppi[:, SIZE_POS].clip(max=psizes_max, min=1)
    padding_mask = x_ppi[:, DIR_POS] == 0 # mask of zero padding
    if ipt_scaler:
        x_ppi[:, IPT_POS] = ipt_scaler.transform(x_ppi[:, IPT_POS].reshape(-1, 1)).reshape(-1) # type: ignore
    if psizes_scaler:
        x_ppi[:, SIZE_POS] = psizes_scaler.transform(x_ppi[:, SIZE_POS].reshape(-1, 1)).reshape(-1) # type: ignore
    x_ppi[padding_mask, IPT_POS] = 0
    x_ppi[padding_mask, SIZE_POS] = 0
    x_ppi = x_ppi.reshape(orig_shape).transpose(0, 2, 1)
    if not use_push_flags:
        x_ppi = x_ppi[:, (IPT_POS, DIR_POS, SIZE_POS), :]
    if zero_ppi_start > 0:
        x_ppi[:,:,:zero_ppi_start] = 0

    if use_packet_histograms:
        x_phist = structured_to_unstructured(x_flowstats[PHISTS_FEATURES], dtype="float32")
        if normalize_packet_histograms:
            src_sizes_pkt_count = x_phist[:, :PHIST_BIN_COUNT].sum(axis=1)[:, np.newaxis]
            dst_sizes_pkt_count = x_phist[:, PHIST_BIN_COUNT:(2*PHIST_BIN_COUNT)].sum(axis=1)[:, np.newaxis]
            np.divide(x_phist[:, :PHIST_BIN_COUNT], src_sizes_pkt_count, out=x_phist[:, :PHIST_BIN_COUNT], where=src_sizes_pkt_count != 0)
            np.divide(x_phist[:, PHIST_BIN_COUNT:(2*PHIST_BIN_COUNT)], dst_sizes_pkt_count, out=x_phist[:, PHIST_BIN_COUNT:(2*PHIST_BIN_COUNT)], where=dst_sizes_pkt_count != 0)
            np.divide(x_phist[:, (2*PHIST_BIN_COUNT):(3*PHIST_BIN_COUNT)], src_sizes_pkt_count - 1, out=x_phist[:, (2*PHIST_BIN_COUNT):(3*PHIST_BIN_COUNT)], where=src_sizes_pkt_count > 1)
            np.divide(x_phist[:, (3*PHIST_BIN_COUNT):(4*PHIST_BIN_COUNT)], dst_sizes_pkt_count - 1, out=x_phist[:, (3*PHIST_BIN_COUNT):(4*PHIST_BIN_COUNT)], where=dst_sizes_pkt_count > 1)
        x_flowstats = structured_to_unstructured(drop_fields(x_flowstats, PHISTS_FEATURES), dtype="float32")
        x_flowstats = np.concatenate([x_flowstats, x_phist], axis=1)
    else:
        x_flowstats = structured_to_unstructured(x_flowstats, dtype="float32")
    np.clip(x_flowstats[:, :len(FLOWSTATS_TO_SCALE)], a_max=flowstats_quantiles, a_min=0, out=x_flowstats[:, :len(FLOWSTATS_TO_SCALE)])
    if flowstats_scaler:
        x_flowstats[:, :len(FLOWSTATS_TO_SCALE)] = flowstats_scaler.transform(x_flowstats[:, :len(FLOWSTATS_TO_SCALE)])

    labels = encoder.transform(np.where(np.isin(labels, known_apps), labels, UNKNOWN_STR_LABEL)).astype("int64") # type: ignore
    if return_torch:
        return torch.from_numpy(x_ppi), torch.from_numpy(x_flowstats), torch.from_numpy(labels)
    return x_ppi, x_flowstats, labels

def init_train_indices(train_data_params: TrainDataParams, servicemap: pd.DataFrame, database_path: str, rng: np.random.RandomState) -> tuple[np.ndarray, np.ndarray, dict[int, str], dict[int, str]]:
    database, train_tables = load_database(database_path, tables_paths=train_data_params.train_tables_paths)
    app_enum = train_tables[0].get_enum(APP_COLUMN)
    all_app_labels = {}
    app_counts = pd.Series(dtype="int64")
    start_time = time.time()
    for i, table_path in enumerate(train_data_params.train_tables_paths):
        all_app_labels[i] = train_tables[i].read(field=APP_COLUMN)
        log.info(f"Reading app column for train table {table_path} took {time.time() - start_time:.2f} seconds"); start_time = time.time()
        app_counts = app_counts.add(pd.Series(all_app_labels[i]).value_counts(), fill_value=0)
    database.close()
    # Handle disabled apps and apps with less than min_samples_per_app samples
    if len(train_data_params.disabled_apps) > 0:
        log.info(f"Disabled applications in dataset config: {sorted(train_data_params.disabled_apps)}")
    disabled_apps_ids = list(map(lambda x: app_enum[x], train_data_params.disabled_apps))
    min_samples_apps_ids = set(app_counts[app_counts<train_data_params.min_train_samples_per_app].index.tolist())
    if len(min_samples_apps_ids) > 0:
        if train_data_params.min_train_samples_check == MinTrainSamplesCheck.WARN_AND_EXIT:
            warnings.warn(f"Found applications with less than {train_data_params.min_train_samples_per_app} train samples: {sorted(map(app_enum, min_samples_apps_ids))}. " +
                            "To disable these applications, add them to config.disabled_apps or set config.min_train_samples_check to disable-apps. To turn off this check, set config.min_train_samples_per_app to zero. Exiting")
            exit()
        elif train_data_params.min_train_samples_check == MinTrainSamplesCheck.DISABLE_APPS:
            log.info(f"Found applications with less than {train_data_params.min_train_samples_per_app} train samples: {sorted(map(app_enum, min_samples_apps_ids))}. " +
                       "Disabling these applications")
            disabled_apps_ids.extend(min_samples_apps_ids)
    # Base indices are indices of samples that are not disabled and have enough samples
    base_indices = {}
    for i, table_path in enumerate(train_data_params.train_tables_paths):
        base_indices[i] = np.nonzero(np.isin(all_app_labels[i], disabled_apps_ids, invert=True))[0]
    base_labels = {table_id: arr[base_indices[table_id]] for table_id, arr in all_app_labels.items()}
    # Apps selection
    if train_data_params.apps_selection != AppSelection.LONGTERM_FIXED:
        app_counts = app_counts[[app for app in app_counts.index.tolist() if app not in disabled_apps_ids]]
        app_counts.index = app_counts.index.map(app_enum)
        app_counts = app_counts.sort_values(ascending=False).astype("int64")
        sorted_apps = app_counts.index.to_list()
        if train_data_params.apps_selection == AppSelection.ALL_KNOWN:
            known_apps = [app for app in sorted_apps if not is_background_app(app)]
            unknown_apps = []
        elif train_data_params.apps_selection == AppSelection.TOPX_KNOWN:
            known_apps, unknown_apps = split_apps_topx_with_provider_groups(sorted_apps=sorted_apps, known_count=train_data_params.apps_selection_topx, servicemap=servicemap)
            if len(known_apps) < train_data_params.apps_selection_topx:
                warnings.warn(f"The number of known applications ({len(known_apps)}) is lower than requested in config.apps_selection_topx ({train_data_params.apps_selection_topx}).")
        elif train_data_params.apps_selection == AppSelection.EXPLICIT_UNKNOWN:
                unknown_apps = train_data_params.apps_selection_explicit_unknown
                missing_unknown_apps = [app for app in unknown_apps if app not in sorted_apps]
                if len(missing_unknown_apps) > 0:
                    raise ValueError(f"Applications configured in config.apps_selection_explicit_unknown are not present in the dataset (or might be disabled): {sorted(missing_unknown_apps)}")
                known_apps = [app for app in sorted_apps if not (is_background_app(app) or app in unknown_apps)]
        else: assert_never(train_data_params.apps_selection)

        log.info(f"Selected {len(known_apps)} known applications and {len(unknown_apps)} unknown applications")
        known_apps_database_enum: dict[int, str] = {int(app_enum[app]): app for app in known_apps}
        unknown_apps_database_enum: dict[int, str] = {int(app_enum[app]): app for app in unknown_apps}
    else:
        assert train_data_params.apps_selection_fixed_longterm is not None
        known_apps_database_enum, unknown_apps_database_enum = train_data_params.apps_selection_fixed_longterm
    known_apps_ids = list(known_apps_database_enum)
    unknown_apps_ids = list(unknown_apps_database_enum)

    train_known_indices, train_unknown_indices = convert_dict_indices(base_indices=base_indices, base_labels=base_labels, known_apps_ids=known_apps_ids, unknown_apps_ids=unknown_apps_ids)
    rng.shuffle(train_known_indices)
    rng.shuffle(train_unknown_indices)
    log.info(f"Processing train indices took {time.time() - start_time:.2f} seconds"); start_time = time.time()
    return train_known_indices, train_unknown_indices, known_apps_database_enum, unknown_apps_database_enum

def init_test_indices(test_data_params: TestDataParams, database_path: str, rng: np.random.RandomState) -> tuple[np.ndarray, np.ndarray]:
    database, test_tables = load_database(database_path, tables_paths=test_data_params.test_tables_paths)
    base_labels = {}
    base_indices = {}
    start_time = time.time()
    for i, table_path in enumerate(test_data_params.test_tables_paths):
        base_labels[i] = test_tables[i].read(field=APP_COLUMN)
        log.info(f"Reading app column for test table {table_path} took {time.time() - start_time:.2f} seconds"); start_time = time.time()
        base_indices[i] = np.arange(len(test_tables[i]))
    database.close()
    known_apps_ids = list(test_data_params.known_apps_database_enum)
    unknown_apps_ids = list(test_data_params.unknown_apps_database_enum)
    test_known_indices, test_unknown_indices = convert_dict_indices(base_indices=base_indices, base_labels=base_labels, known_apps_ids=known_apps_ids, unknown_apps_ids=unknown_apps_ids)
    rng.shuffle(test_known_indices)
    rng.shuffle(test_unknown_indices)
    log.info(f"Processing test indices took {time.time() - start_time:.2f} seconds"); start_time = time.time()
    return test_known_indices, test_unknown_indices

def fit_or_load_scalers(dataset_config: DatasetConfig, train_indices: np.ndarray) -> tuple[Scaler, pd.Series, Scaler, Scaler]:
    train_data_path = dataset_config._get_train_data_path()
    flowstats_scaler_path = os.path.join(train_data_path, "stand", f"flowstats_scaler-{dataset_config.flowstats_scaler}-q{dataset_config.flowstats_clip}.pickle")
    flowstats_quantiles_path = os.path.join(train_data_path, "stand", f"flowstats_quantiles-q{dataset_config.flowstats_clip}.pickle")
    ipt_scaler_path = os.path.join(train_data_path, "stand", f"ipt_scaler-{dataset_config.ipt_scaler}-ipt_min{dataset_config.ipt_min}-ipt_max{dataset_config.ipt_max}.pickle")
    psizes_sizes_scaler_path = os.path.join(train_data_path, "stand", f"psizes_scaler-{dataset_config.psizes_scaler}-psizes_max{dataset_config.psizes_max}.pickle")
    if os.path.isfile(flowstats_scaler_path) and os.path.isfile(flowstats_quantiles_path) and os.path.isfile(ipt_scaler_path) and os.path.isfile(psizes_sizes_scaler_path):
        flowstats_scaler = pickle_load(flowstats_scaler_path)
        flowstats_quantiles = pickle_load(flowstats_quantiles_path)
        ipt_scaler = pickle_load(ipt_scaler_path)
        psizes_scaler = pickle_load(psizes_sizes_scaler_path)
    else:
        if dataset_config.flowstats_scaler == ScalerEnum.ROBUST:
            flowstats_scaler = RobustScaler()
        elif dataset_config.flowstats_scaler == ScalerEnum.STANDARD:
            flowstats_scaler = StandardScaler()
        elif dataset_config.flowstats_scaler == ScalerEnum.MINMAX:
            flowstats_scaler = MinMaxScaler()
        elif dataset_config.flowstats_scaler == ScalerEnum.NO_SCALER:
            flowstats_scaler = None
        else: assert_never(dataset_config.flowstats_scaler)

        if dataset_config.ipt_scaler == ScalerEnum.ROBUST:
            ipt_scaler = RobustScaler()
        elif dataset_config.ipt_scaler == ScalerEnum.STANDARD:
            ipt_scaler = StandardScaler()
        elif dataset_config.ipt_scaler == ScalerEnum.MINMAX:
            ipt_scaler = MinMaxScaler()
        elif dataset_config.ipt_scaler == ScalerEnum.NO_SCALER:
            ipt_scaler = None
        else: assert_never(dataset_config.ipt_scaler)

        if dataset_config.psizes_scaler == ScalerEnum.ROBUST:
            psizes_scaler = RobustScaler()
        elif dataset_config.psizes_scaler == ScalerEnum.STANDARD:
            psizes_scaler = StandardScaler()
        elif dataset_config.psizes_scaler == ScalerEnum.MINMAX:
            psizes_scaler = MinMaxScaler()
        elif dataset_config.psizes_scaler == ScalerEnum.NO_SCALER:
            psizes_scaler = None
        else: assert_never(dataset_config.psizes_scaler)

        if isinstance(dataset_config.fit_scalers_samples, int) and dataset_config.fit_scalers_samples > len(train_indices):
            warnings.warn(f"The number of samples for fitting scalers ({dataset_config.fit_scalers_samples}) is larger than the number of train samples ({len(train_indices)}), using the number of train samples instead")
            dataset_config.fit_scalers_samples = len(train_indices)
        fit_scalers_rng = get_fresh_random_generator(dataset_config=dataset_config, section=RandomizedSection.FIT_SCALERS_SAMPLE)
        if isinstance(dataset_config.fit_scalers_samples, float):
            num_samples = int(dataset_config.fit_scalers_samples * len(train_indices))
        else:
            num_samples = dataset_config.fit_scalers_samples
        fit_scalers_indices = train_indices[fit_scalers_rng.choice(len(train_indices), size=num_samples, replace=False)]
        flowstats_quantiles = fit_scalers(
            database_path=dataset_config.database_path,
            train_tables_paths=dataset_config._get_train_tables_paths(),
            fit_scalers_indices=fit_scalers_indices,
            flowstats_scaler=flowstats_scaler,
            flowstats_quantile_clip=dataset_config.flowstats_clip,
            ipt_scaler=ipt_scaler,
            psizes_scaler=psizes_scaler,
            ipt_min=dataset_config.ipt_min,
            ipt_max=dataset_config.ipt_max,
            psizes_max=dataset_config.psizes_max)
        pickle_dump(flowstats_scaler, flowstats_scaler_path)
        pickle_dump(flowstats_quantiles, flowstats_quantiles_path)
        pickle_dump(ipt_scaler, ipt_scaler_path)
        pickle_dump(psizes_scaler, psizes_sizes_scaler_path)
    return flowstats_scaler, flowstats_quantiles, ipt_scaler, psizes_scaler

def fit_scalers(database_path: str, train_tables_paths: list[str], fit_scalers_indices: np.ndarray, flowstats_scaler, flowstats_quantile_clip: float, ipt_scaler, psizes_scaler, ipt_min: int, ipt_max: int, psizes_max: int) -> pd.Series:
    start_time = time.time()
    database, tables = load_database(database_path, tables_paths=train_tables_paths)
    data = load_data_from_pytables(tables=tables, indices=fit_scalers_indices, data_dtype=tables[0].dtype)
    database.close()
    # PPI
    data_ppi = data[PPI_COLUMN].astype("float32")
    ppi_channels = data_ppi.shape[1]
    data_ppi = data_ppi.transpose(0, 2, 1).reshape(-1, ppi_channels)
    padding_mask = data_ppi[:, DIR_POS] == 0 # mask of padded packets
    if ipt_scaler:
        train_ipt = data_ppi[:, IPT_POS].clip(max=ipt_max, min=ipt_min)
        train_ipt[padding_mask] = np.nan # nans are ignored in sklearn scalers
        if isinstance(ipt_scaler, MinMaxScaler):
            # let zero be the minimum for minmax scaling
            train_ipt = np.concatenate((train_ipt, [0]))
        ipt_scaler.fit(train_ipt.reshape(-1, 1))
    if psizes_scaler:
        train_psizes = data_ppi[:, SIZE_POS].clip(max=psizes_max, min=1)
        train_psizes[padding_mask] = np.nan
        if isinstance(psizes_scaler, MinMaxScaler):
            train_psizes = np.concatenate((train_psizes, [0]))
        psizes_scaler.fit(train_psizes.reshape(-1, 1))
    # FLOWSTATS
    train_flowstats = pd.DataFrame(data, columns=FLOWSTATS_TO_SCALE)
    upper_quantiles = train_flowstats.quantile(flowstats_quantile_clip)
    upper_quantiles[FLOWSTATS_NO_CLIP] = np.Inf # disable clipping for features with "fixed" range
    if flowstats_scaler:
        train_flowstats = train_flowstats.clip(upper=upper_quantiles, lower=0, axis=1).to_numpy() # type: ignore
        flowstats_scaler.fit(train_flowstats)
    log.info(f"Reading data and fitting scalers took {time.time() - start_time:.2f} seconds")
    return upper_quantiles

def load_database(database_path: str, tables_paths: Optional[list[str]] = None, mode: str = "r") -> tuple[tb.File, dict[int, Any]]: # dict[int, tb.Table]
    database = tb.open_file(database_path, mode=mode)
    if tables_paths is None:
        tables_paths = list(map(lambda x: x._v_pathname, iter(database.get_node(f"/flows"))))
    tables = {}
    try:
        for i, table_path in enumerate(tables_paths):
            tables[i] = database.get_node(table_path)
    except tb.NoSuchNodeError as e:
        raise e
    return database, tables

def list_all_tables(database_path: str) -> list[str]:
    with tb.open_file(database_path, mode="r") as database:
        return list(map(lambda x: x._v_pathname, iter(database.get_node(f"/flows"))))

def convert_dict_indices(base_indices: dict[int, np.ndarray], base_labels: dict[int, np.ndarray], known_apps_ids: list[int], unknown_apps_ids: list[int]) -> tuple[np.ndarray, np.ndarray]:
    is_known = {table_id: np.isin(table_arr, known_apps_ids) for table_id, table_arr in base_labels.items()}
    is_unknown = {table_id: np.isin(table_arr, unknown_apps_ids) for table_id, table_arr in base_labels.items()}
    known_indices_dict = {table_id: table_arr[is_known[table_id]] for table_id, table_arr in base_indices.items()}
    unknown_indices_dict = {table_id: table_arr[is_unknown[table_id]] for table_id, table_arr in base_indices.items()}
    known_labels_dict = {table_id: table_arr[is_known[table_id]] for table_id, table_arr in base_labels.items()}
    unknown_labels_dict = {table_id: table_arr[is_unknown[table_id]] for table_id, table_arr in base_labels.items()}
    known_indices = np.column_stack((
        np.concatenate([[table_id] * table_arr.sum() for table_id, table_arr in is_known.items()]),
        np.concatenate(list(known_indices_dict.values())),
        np.concatenate(list(known_labels_dict.values()))))
    unknown_indices = np.column_stack((
        np.concatenate([[table_id] * table_arr.sum() for table_id, table_arr in is_unknown.items()]),
        np.concatenate(list(unknown_indices_dict.values())),
        np.concatenate(list(unknown_labels_dict.values()))))
    return known_indices, unknown_indices

def load_data_from_pytables(tables, indices: np.ndarray, data_dtype: np.dtype) -> np.ndarray:
    sorted_indices = indices[indices[:, INDICES_TABLE_POS].argsort(kind="stable")]
    unique_tables, split_bounderies = np.unique(sorted_indices[:, INDICES_TABLE_POS], return_index=True)
    indices_per_table = np.split(sorted_indices, split_bounderies[1:])
    data = np.empty(len(indices), dtype=data_dtype)
    for table_id, table_indices in zip(unique_tables, indices_per_table):
        data[np.where(indices[:, INDICES_TABLE_POS] == table_id)[0]] = tables[table_id].read_coordinates(table_indices[:, INDICES_INDEX_POS])
    return data



import logging
import os
from collections import Counter
from typing import Any, Literal

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from numpy.lib.recfunctions import structured_to_unstructured
from torch.utils.data import BatchSampler, DataLoader, SequentialSampler
from tqdm import tqdm

from cesnet_datazoo.config import Protocol
from cesnet_datazoo.constants import (APP_COLUMN, CATEGORY_COLUMN, FLOWEND_REASON_FEATURES,
                                      PHISTS_FEATURES, PPI_COLUMN, SIZE_POS)
from cesnet_datazoo.pytables_data.indices_setup import sort_indices
from cesnet_datazoo.pytables_data.pytables_dataset import (PyTablesDataset, list_all_tables,
                                                           load_database, worker_init_fn)


def pick_quic_fields(batch):
    return (
        batch["QUIC_SNI"],
        batch["QUIC_USERAGENT"],
        batch["QUIC_VERSION"],
    )

def pick_stats_fields(batch):
    return (
        batch[PPI_COLUMN],
        batch["DURATION"],
        batch["PACKETS"] + batch["PACKETS_REV"],
        batch["BYTES"] + batch["BYTES_REV"],
        batch[APP_COLUMN],
        batch[CATEGORY_COLUMN],
    )

def pick_extra_fields(batch, flowstats_features: list[str]):
    return (
        batch["DST_ASN"],
        batch[PHISTS_FEATURES],
        batch[[f for f in FLOWEND_REASON_FEATURES if f in flowstats_features]],
    )

def simple_collate_fn(batch):
    return batch

def compute_dataset_statistics(database_path: str, output_dir: str, flowstats_features: list[str], protocol: Protocol, extra_fields: bool, disabled_apps: list[str], num_samples: int | Literal["all"] = 10_000_000, num_workers: int = 4, batch_size: int = 4096, silent: bool = False):
    stats_pdf_path = os.path.join(output_dir, "dataset-statistics.pdf")
    stats_csv_path = os.path.join(output_dir, "dataset-statistics.csv")
    categories_csv_path = os.path.join(output_dir, "categories.csv")
    categories_figure_path = os.path.join(output_dir, "categories-tikzpicture.txt")
    app_path = os.path.join(output_dir, "app.csv")
    asn_path = os.path.join(output_dir, "asn.csv")
    phist_path = os.path.join(output_dir, "phists.csv")
    packet_sizes_path = os.path.join(output_dir, "ppi-packet-sizes.csv")
    flow_endreason_path = os.path.join(output_dir, "flow-endreason.csv")
    quic_sni_path = os.path.join(output_dir, "quic-sni.csv")
    quic_ua_path = os.path.join(output_dir, "quic-ua.csv")
    quic_version_path = os.path.join(output_dir, "quic-version.csv")

    df_categories = pd.DataFrame()
    df_phist = pd.DataFrame()
    app_series = pd.Series(dtype="int64")
    quic_sni_series = pd.Series(dtype="int64")
    quic_ua_series = pd.Series(dtype="int64")
    quic_version_series = pd.Series(dtype="int64")
    asn_series = pd.Series(dtype="int64")
    flow_endreason_series = pd.Series(dtype="int64")
    feature_duration = []
    feature_packets_total = []
    feature_bytes_total = []
    packet_sizes_counter = Counter()
    if not silent:
        print(f"Reading data from {database_path} for statistics")
    table_paths = list_all_tables(database_path)
    stats_dataset = PyTablesDataset(database_path=database_path, tables_paths=table_paths, flowstats_features=flowstats_features, disabled_apps=disabled_apps, indices=None, return_all_fields=True)
    if num_samples != "all":
        subset_indices = np.random.randint(low=0, high=len(stats_dataset.indices), size=num_samples)
        stats_dataset.indices = sort_indices(stats_dataset.indices[subset_indices])
    stats_batch_sampler = BatchSampler(sampler=SequentialSampler(stats_dataset), batch_size=batch_size, drop_last=False)
    stats_dloader = DataLoader(
        stats_dataset,
        pin_memory=False,
        num_workers=num_workers,
        worker_init_fn=worker_init_fn,
        collate_fn=simple_collate_fn,
        persistent_workers=False,
        batch_size=None,
        sampler=stats_batch_sampler)
    if num_workers == 0:
        stats_dataset.pytables_worker_init()

    for batch, batch_idx in tqdm(stats_dloader, total=len(stats_dloader), disable=silent):
        ppi, duration, packets_total, bytes_total, app, cat = pick_stats_fields(batch)
        # Saving feature values for distribution plots
        feature_duration.append(duration)
        feature_packets_total.append(packets_total)
        feature_bytes_total.append(bytes_total)
        packet_sizes_counter.update(ppi[:, SIZE_POS, :].flatten())
        # Aggregating features for value_counts
        app_series = app_series.add(pd.Series(app).value_counts(), fill_value=0)
        # Grouping features per categories
        df1 = pd.DataFrame(data={"cat": cat, "BYTES_TOTAL": bytes_total})
        flow_counts = df1["cat"].value_counts().rename("FLOW_COUNT")
        byte_volumes = df1.groupby("cat")["BYTES_TOTAL"].sum()
        df_categories = df_categories.add(pd.concat((flow_counts, byte_volumes), axis=1), fill_value=0)
        # QUIC features
        if protocol == Protocol.QUIC:
            sni, user_agent, quic_version = pick_quic_fields(batch)
            quic_sni_series = quic_sni_series.add(pd.Series(sni).str.decode("utf-8").value_counts(), fill_value=0)
            quic_ua_series = quic_ua_series.add(pd.Series(user_agent).str.decode("utf-8").value_counts(), fill_value=0)
            quic_version_series = quic_version_series.add(pd.Series(quic_version).value_counts(), fill_value=0)
        if extra_fields:
            asn, phist, flowend_reason = pick_extra_fields(batch, flowstats_features=flowstats_features)
            asn_series = asn_series.add(pd.Series(asn).value_counts(), fill_value=0)
            flow_endreason_series = flow_endreason_series.add(pd.Series(structured_to_unstructured(flowend_reason).sum(axis=0)), fill_value=0)
            df2 = pd.DataFrame(data=zip(*np.split(structured_to_unstructured(phist).sum(axis=0), 4)), columns=PHISTS_FEATURES)
            df_phist = df_phist.add(df2, fill_value=0)
    feature_duration = np.concatenate(feature_duration)
    feature_packets_total = np.concatenate(feature_packets_total)
    feature_bytes_total = np.concatenate(feature_bytes_total)
    packet_sizes_counter.pop(0) # remove zero packets as those are just padding
    assert df_categories["BYTES_TOTAL"].sum() == feature_bytes_total.sum()
    assert df_categories["FLOW_COUNT"].sum() == app_series.sum()

    # Flow statistics distribution output
    df_flowstats = pd.DataFrame(data={"FLOW DURATION": feature_duration, "FLOW BYTE VOLUME": feature_bytes_total, "FLOW LENGTH": feature_packets_total}).describe()
    df_flowstats.to_csv(stats_csv_path)
    # Categories tikzpicture and csv output; first, get the categories and applications enum
    temp_database, temp_tables = load_database(database_path=database_path, tables_paths=table_paths[:1])
    cat_enum = temp_tables[0].get_enum(CATEGORY_COLUMN)
    app_enum = temp_tables[0].get_enum(APP_COLUMN)
    temp_database.close()
    df_categories.index = df_categories.index.map(cat_enum)
    df_categories = df_categories.drop("default", errors="ignore")
    df_categories["FLOW_PERC"] = df_categories["FLOW_COUNT"] / sum(df_categories["FLOW_COUNT"]) * 100
    df_categories["BYTES_PERC"] = df_categories["BYTES_TOTAL"] / sum(df_categories["BYTES_TOTAL"]) * 100
    df_categories = df_categories.round(3)
    df_categories = df_categories.sort_values("FLOW_PERC", ascending=False)
    with open(categories_figure_path, "w") as f:
        print(f"Categories:\n{df_categories.index.tolist()}", file=f)
        print(f"Flow Percentage:\n{list(df_categories['FLOW_PERC'].items())}", file=f)
        print(f"Byte Volume Percentage:\n{list(df_categories['BYTES_PERC'].items())}", file=f)
    df_categories.rename({"FLOW_COUNT": "FLOW COUNT", "BYTES_TOTAL": "BYTE VOLUME", "FLOW_PERC": "FLOW PERC", "BYTES_PERC": "BYTE VOLUME PERC"}, axis=1).to_csv(categories_csv_path, index_label="CATEGORIES")
    # Application distribution output
    app_df = pd.DataFrame({"COUNT": app_series.sort_values(ascending=False).astype("int64")})
    app_df["PERC"] = (app_df["COUNT"] / app_df["COUNT"].sum() * 100).round(3)
    app_df.index = app_df.index.map(app_enum)
    app_df.index.name = "LABEL"
    app_df.to_csv(app_path)
    # Packet sizes histogram output
    packet_sizes_df = pd.DataFrame({"COUNT": pd.Series(packet_sizes_counter)}).sort_index()
    packet_sizes_df["PERC"] = (packet_sizes_df["COUNT"] / packet_sizes_df["COUNT"].sum() * 100).round(3)
    packet_sizes_df.index.name = "PACKET SIZE"
    packet_sizes_df.to_csv(packet_sizes_path)
    # QUIC features output
    if protocol == Protocol.QUIC:
        quic_ua_series = quic_ua_series.rename(index={"": "No User Agent"})
        quic_ua_series.sort_values(ascending=False).astype("int64").rename("COUNT").to_csv(quic_ua_path, index_label="QUIC USER AGENT")
        quic_sni_series.sort_values(ascending=False).astype("int64").rename("COUNT").to_csv(quic_sni_path, index_label="QUIC SNI")
        quic_version_df = pd.DataFrame({"COUNT": quic_version_series.sort_values(ascending=False).astype("int64")})
        quic_version_df["PERC"] = (quic_version_df["COUNT"] / quic_version_df["COUNT"].sum() * 100).round(3)
        quic_version_df.index = quic_version_df.index.map(hex)
        quic_version_df.index.name = "QUIC VERSION"
        quic_version_df.to_csv(quic_version_path)
    if extra_fields:
        # ASN distribution output
        asn_df = pd.DataFrame({"COUNT": asn_series.sort_values(ascending=False).astype("int64")})
        asn_df["PERC"] = (asn_df["COUNT"] / asn_df["COUNT"].sum() * 100).round(3)
        asn_df.index.name = "DESTINATION ASN"
        asn_df.to_csv(asn_path)
        # Flow end reason output
        flow_endreason_df = pd.DataFrame({"COUNT": flow_endreason_series.astype("int64")})
        flow_endreason_df["PERC"] = (flow_endreason_df["COUNT"] / flow_endreason_df["COUNT"].sum() * 100).round(3)
        flow_endreason_df.index.name = "FLOW ENDREASON"
        flow_endreason_df.index = pd.Index([f for f in FLOWEND_REASON_FEATURES if f in flowstats_features])
        flow_endreason_df.to_csv(flow_endreason_path)
        # PHIST output
        df_phist.index.name = "BINS"
        df_phist.columns = list(map(lambda x: x.upper().replace("_", " "), PHISTS_FEATURES))
        df_phist = df_phist.astype("int64")
        for i, column in zip((1, 3, 5, 7), df_phist.columns):
            df_phist.insert(i, column + " PERC", (df_phist[column] / df_phist[column].sum() * 100).round(3))
        df_phist.to_csv(phist_path)

    # Dataset stats figure
    axes: Any
    fig, axes = plt.subplots(2, 2, figsize=(7, 7))
    ax1 = axes[0][0]
    ax2 = axes[0][1]
    ax3 = axes[1][0]
    ax4 = axes[1][1]
    x_log_text_margin = 1.3
    x_text_margin = 1.7
    y_prob_text_margin = 0.03
    x_psizes_margin = 0.5

    sns.ecdfplot(feature_duration, ax=ax1, log_scale=True)
    cdf = ax1.get_lines()[0].get_data()
    ax1.set_title("Flow Duration")
    ax1.set_xlabel("Seconds")
    prob_at_1s = cdf[1][np.searchsorted(cdf[0], 1, side="right")]
    ax1.axvline(1, color="grey", linestyle="--")
    ax1.text(1 * x_log_text_margin, y_prob_text_margin, f"{prob_at_1s*100:.1f}% < 1s", rotation=90)
    prob_at_30s = cdf[1][np.searchsorted(cdf[0], 30, side="right")]
    ax1.axvline(30, color="grey", linestyle="--")
    ax1.text(30 * x_log_text_margin, y_prob_text_margin, f"{prob_at_30s*100:.1f}% < 30s", rotation=90)

    sns.ecdfplot(feature_bytes_total, ax=ax2, log_scale=True)
    cdf = ax2.get_lines()[0].get_data()
    ax2.set_title("Flow Byte Volume")
    ax2.set_xlabel("Bytes")
    prob_at_10kb = cdf[1][np.searchsorted(cdf[0], 10_000, side="right")]
    ax2.axvline(10_000, color="grey", linestyle="--")
    ax2.text(10_000 * x_log_text_margin, y_prob_text_margin, f"{prob_at_10kb*100:.1f}% < 10KB", rotation=90)
    prob_at_1mb = cdf[1][np.searchsorted(cdf[0], 1_000_000, side="right")]
    ax2.axvline(1_000_000, color="grey", linestyle="--")
    ax2.text(1_000_000 * x_log_text_margin, y_prob_text_margin, f"{prob_at_1mb*100:.1f}% < 1MB", rotation=90)
    mb_at_999perc = cdf[0][np.searchsorted(cdf[1], 0.999, side="right")]
    ax2.axvline(mb_at_999perc, color="grey", linestyle="--")
    ax2.text(mb_at_999perc * x_log_text_margin, y_prob_text_margin, f"99.9% < {mb_at_999perc / 1_000_000:.1f}1MB", rotation=90)

    feature_packets_total = np.clip(feature_packets_total, a_max=100, a_min=0)
    sns.ecdfplot(feature_packets_total, ax=ax3)
    cdf = ax3.get_lines()[0].get_data()
    ax3.set_title("Flow Length")
    ax3.set_xlabel("Packets")
    ax3.set_xlim(0, 110)
    prob_at_30packets = cdf[1][np.searchsorted(cdf[0], 30, side="left")]
    prob_at_100packets = cdf[1][np.searchsorted(cdf[0], 100, side="left")]
    ax3.axvline(30, color="grey", linestyle="--")
    ax3.text(30 + x_text_margin, y_prob_text_margin, f"{prob_at_30packets*100:.1f}% < 30 packets", rotation=90)
    ax3.axvline(100, color="grey", linestyle="--")
    ax3.text(100 + x_text_margin, y_prob_text_margin, f"{prob_at_100packets*100:.1f}% < 100 packets", rotation=90)

    packet_sizes_hist = list(zip(*sorted(dict(packet_sizes_counter).items(), key=lambda x: x[0])))
    sns.histplot(x=packet_sizes_hist[0], weights=packet_sizes_hist[1], ax=ax4, binwidth=50, binrange=(0, 1500), stat="proportion")
    ax4.set_title("Packet Sizes")
    ax4.set_xlabel("Bytes")

    plt.tight_layout()
    fig.savefig(stats_pdf_path, bbox_inches="tight")



import numpy as np
from sklearn.metrics import accuracy_score, precision_recall_fscore_support

from cesnet_datazoo.metrics.provider_metrics import (per_app_provider_metrics,
                                                       provider_accuracies)
from cesnet_datazoo.utils.class_info import ClassInfo


def better_classification_report(y_true: np.ndarray, y_pred: np.ndarray, cm: np.ndarray, labels: list[int], class_info: ClassInfo, digits: int = 2, zero_division: int = 0) -> tuple[str, dict[str, float]]:
    p, r, f1, s  = precision_recall_fscore_support(y_true, y_pred,
                                                    labels=labels,
                                                    zero_division=zero_division)
    sc_p, sc_r, sc_f1 = per_app_provider_metrics(cm, class_info=class_info)
    predicted_unknown = cm[:, -1]
    with np.errstate(divide="ignore", invalid="ignore"):
        predicted_unknown_perc = predicted_unknown / s
        predicted_unknown_perc = np.nan_to_num(predicted_unknown_perc)
    headers = ["precision (pr)", "recall (pr)", "f1-score (pr)", "pred unknown", "support"]
    headers_fmt = "{:>{width}} {:>15} {:>15} {:>15} {:>15} {:>9}\n"
    width = max(max(len(cn) for cn in class_info.target_names), len("failed provider acc"))
    report = headers_fmt.format("", *headers, width=width)
    report += "\n"
    row_fmt_provider = "{:>{width}} " + 3 * " {:>7.{digits}f} ({:.{digits}f}) " + " {:>7} ({:.{digits}f}) " + "{:>9}\n"
    row_fmt = "{:>{width}} " + 3 * " {:>7.{digits}f}        " + " {:>7} ({:.{digits}f}) " + "{:>9}\n"
    rows = zip(map(class_info.target_names.__getitem__, labels), p, sc_p, r, sc_r, f1, sc_f1, predicted_unknown, predicted_unknown_perc, s) # type: ignore
    for row in rows:
        app, p_, _, r_, _, f1_, _, u_, up_, s_ = row
        if class_info.has_provider[app]:
            report += row_fmt_provider.format(*row, width=width, digits=digits)
        else:
            report += row_fmt.format(app, p_, r_, f1_, u_, up_, s_, width=width, digits=digits)
    report += "\n"

    # Computing averages, ignoring the last element with metrics for the unknown class
    samples_sum = s.sum() # type: ignore
    predicted_unknown_sum = predicted_unknown.sum()
    avg_p, avg_r, avg_f1 = np.average(p[:-1]), np.average(r[:-1]), np.average(f1[:-1]) # type: ignore
    avg_sc_p = np.average(np.where(np.isnan([np.nan if x is None else x for x in sc_p]), p, sc_p)[:-1])
    avg_sc_r = np.average(np.where(np.isnan([np.nan if x is None else x for x in sc_r]), r, sc_r)[:-1])
    avg_sc_f1 = np.average(np.where(np.isnan([np.nan if x is None else x for x in sc_f1]), f1, sc_f1)[:-1])
    row_avg = [avg_p, avg_sc_p, avg_r, avg_sc_r, avg_f1, avg_sc_f1, predicted_unknown_sum, samples_sum]

    headers_avg = ["precision (pr)", "recall (pr)", "f1-score (pr)", "pred unknown", "support"]
    row_fmt_avg = "{:>{width}} " + 3 * " {:>6.{digits}} ({:.{digits}f}) " + "{:>15} " + "{:>9}\n"
    digits = 3 # show more precise averages
    report += headers_fmt.format("", *headers_avg, width=width)
    report += row_fmt_avg.format("macro avg", *row_avg, width=width, digits=digits)

    acc = accuracy_score(y_true, y_pred)
    provider_acc, failed_provider_acc = provider_accuracies(y_true, y_pred, class_info=class_info)

    row_fmt_acc = "{:>{width}} {:>15} {:>15} {:>7.{digits}f}\n"
    report += row_fmt_acc.format("acc", "", "", acc, width=width, digits=digits)
    report += row_fmt_acc.format("provider acc", "", "", provider_acc, width=width, digits=digits)
    report += row_fmt_acc.format("failed provider acc", "", "", failed_provider_acc, width=width, digits=digits)
    metrics = {
        "Test/Accuracy": acc,
        "Test/Provider Accuracy": provider_acc,
        "Test/Failed Provider Accuracy": failed_provider_acc,
        "Test/Fscore": avg_f1,
        "Test/Provider Fscore": avg_sc_f1,
        "Test/Recall": avg_r,
        "Test/Provider Recall": avg_sc_r,
    }
    return report, metrics
